#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012, 2013 DResearch Fahrzeugelektronik GmbH
# Maintained by project-monk@dresearch-fe.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This is the package overview of :term:`MONK`. If anything is unclear, you might
have a look into :ref:`chap-intro`.

The following texts describe the three layers that were explained in
:ref:`intro-layers`:
"""

__title__ = "monk_tf"
__version__ = "0.8.3"
__author__ = "DResearch Fahrzeugelektronik GmbH"
__license__ = "GPLv3+"
__copyright__ = "Copyright 2012, 2013, 2014 DResearch Fahrzeugelektronik GmbH"
